from numba.core import types
from numba.extending import overload


def maybe_swap(a, i, j):
    """
    Swap the elements in two positions in an array.
    In the special case that the array is ``None``, it does nothing.
    """
    if a is not None:
        a[i], a[j] = a[j], a[i]


@overload(maybe_swap)
def _ovl_swap(a, i, j):
    def swap(a, i, j):
        a[i], a[j] = a[j], a[i]

    def nothing(a, i, j):
        pass

    if isinstance(a, types.NoneType):
        return nothing
    else:
        return swap



from csr import CSR, constructors
from csr.test_utils import has_memory
import numpy as np

import pytest
from hypothesis import given
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph


@given(st.integers(0, 1000), st.integers(0, 1000))
def test_empty(nrows, ncols):
    csr = CSR.empty(nrows, ncols)
    assert csr.nrows == nrows
    assert csr.ncols == ncols
    assert csr.nnz == 0
    assert all(csr.rowptrs == 0)
    assert len(csr.rowptrs) == nrows + 1
    assert len(csr.colinds) == 0


@given(st.data(), st.integers(0, 1000), st.integers(0, 1000),
       st.one_of(st.booleans(), st.just('f4'), st.just('f8')))
def test_empty_csr(data, nrows, ncols, vdt):
    sizes = data.draw(nph.arrays(np.int32, nrows, elements=st.integers(0, ncols)))
    csr = CSR.empty(nrows, ncols, sizes, values=vdt)
    assert csr.nrows == nrows
    assert csr.ncols == ncols
    assert csr.nnz == np.sum(sizes)
    assert len(csr.rowptrs) == nrows + 1
    assert csr.rowptrs.dtype == np.int32
    assert all(csr.row_nnzs() == sizes)
    assert len(csr.colinds) == np.sum(sizes)
    if vdt:
        assert csr.values is not None
        if vdt is not True:
            assert csr.values.dtype == vdt
        assert csr.values.shape == (csr.nnz,)
    else:
        assert csr.values is None


@given(st.data(), st.integers(0, 1000), st.integers(0, 1000))
def test_create_from_sizes(data, nrows, ncols):
    sizes = data.draw(nph.arrays(np.int32, nrows, elements=st.integers(0, ncols)))
    csr = constructors.create_from_sizes(nrows, ncols, sizes)
    assert csr.nrows == nrows
    assert csr.ncols == ncols
    assert csr.nnz == np.sum(sizes)
    assert len(csr.rowptrs) == nrows + 1
    assert csr.rowptrs.dtype == np.int32
    assert all(csr.row_nnzs() == sizes)
    assert len(csr.colinds) == np.sum(sizes)


@pytest.mark.skipif(not has_memory(12), reason='insufficient memory')
def test_large_init():
    # 10M * 500 = 5B >= INT_MAX
    nrows = 10000000
    ncols = 500
    nnz = nrows * 250

    rowptrs = np.arange(0, nnz + 1, 250, dtype=np.int64)
    assert len(rowptrs) == nrows + 1
    assert rowptrs[-1] == nnz

    try:
        colinds = np.empty(nnz, dtype=np.intc)
    except MemoryError:
        pytest.skip('insufficient memory')

    csr = CSR(nrows, ncols, nnz, rowptrs, colinds, None)
    assert csr.nrows == nrows
    assert csr.ncols == ncols
    assert csr.nnz == nnz
    assert csr.rowptrs.dtype == np.dtype('i8')


@pytest.mark.skipif(not has_memory(12), reason='insufficient memory')
def test_large_empty():
    # 10M * 250 = 2.5B >= INT_MAX
    nrows = 10000000
    ncols = 500
    nnz = nrows * 250

    row_nnzs = np.full(nrows, 250, dtype='i4')

    try:
        csr = CSR.empty(nrows, ncols, row_nnzs=row_nnzs, values=False)
    except MemoryError:
        pytest.skip('insufficient memory')

    assert csr.nrows == nrows
    assert csr.ncols == ncols
    assert csr.nnz == nnz
    assert csr.rowptrs.dtype == np.dtype('i8')
    assert np.all(csr.rowptrs >= 0)
    assert np.all(np.diff(csr.rowptrs) == 250)



"""
Routines for working with matrix structure.
"""

import numpy as np
from numba import njit
from .csr import CSR
from . import _util
from .constructors import create, create_novalues

@njit(nogil=True)
def _from_coo_structure(nrows, rows, cols):
    "Transform COO structure into CSR structure"
    nnz = len(rows)
    counts = np.zeros(nrows, dtype=np.int64)
    for r in rows:
        counts[r] += 1

    rowptrs = np.zeros(nrows + 1, dtype=np.int64)
    for i in range(nrows):
        rowptrs[i + 1] = rowptrs[i] + counts[i]

    rpos = rowptrs.copy()
    out_cols = np.empty_like(cols)

    for i in range(nnz):
        row = rows[i]
        pos = rpos[row]
        out_cols[pos] = cols[i]
        rpos[row] += 1

    return rowptrs, out_cols


@njit(nogil=True)
def _from_coo_values(nrows, rows, cols, values):
    "Transform COO w/ values into CSR"
    nnz = len(rows)
    counts = np.zeros(nrows, dtype=np.int64)
    for r in rows:
        counts[r] += 1

    rowptrs = np.zeros(nrows + 1, dtype=np.int64)
    for i in range(nrows):
        rowptrs[i + 1] = rowptrs[i] + counts[i]

    rpos = rowptrs.copy()
    out_cols = np.empty_like(cols)
    out_vals = np.empty_like(values)

    for i in range(nnz):
        row = rows[i]
        pos = rpos[row]
        out_cols[pos] = cols[i]
        out_vals[pos] = values[i]
        rpos[row] += 1

    return rowptrs, out_cols, out_vals


def from_coo(nrows, rows, cols, values=None):
    if values is None:
        rps, cols = _from_coo_structure(nrows, rows, cols)
        vals = None
    else:
        rps, cols, vals = _from_coo_values(nrows, rows, cols, values)
    return rps, cols, vals


def subset_rows(csr, begin, end):
    "Take a subset of the rows of a CSR."
    st = csr.rowptrs[begin]
    ed = csr.rowptrs[end]
    rps = csr.rowptrs[begin:(end + 1)] - st

    cis = csr.colinds[st:ed]
    if csr.values is not None:
        vs = csr.values[st:ed]
    else:
        vs = None
    return CSR(end - begin, csr.ncols, ed - st, rps, cis, vs)


@njit(nogil=True)
def _pick_rows_nvs(csr: CSR, rows: np.ndarray):
    # how many values do we need
    # this is equivalent to np.sum(tag_mat.row_nnzs()[items]) (which would be fast in Python),
    # but uses less memory
    nr = len(rows)
    nnz = 0
    for i in rows:
        sp, ep = csr.row_extent(i)
        rl = ep - sp
        nnz += rl

    # allocate result arrays
    rowptrs = np.empty(nr + 1, dtype=np.int32)
    colinds = np.empty(nnz, dtype=np.int32)
    pos = 0

    # copy each item's rows to result array
    for ii, i in enumerate(rows):
        # get row start/end
        sp, ep = csr.row_extent(i)
        # how many tags for this item?
        itc = ep - sp
        end = pos + itc
        # copy values
        colinds[pos:end] = csr.colinds[sp:ep]
        # set offset
        rowptrs[ii] = pos
        # update position for storing results
        pos = end
    rowptrs[nr] = pos

    return CSR(nr, csr.ncols, nnz, rowptrs, colinds, None)


@njit(nogil=True)
def _pick_rows(csr: CSR, rows: np.ndarray):
    # how many values do we need
    # this is equivalent to np.sum(tag_mat.row_nnzs()[items]) (which would be fast in Python),
    # but uses less memory
    nr = len(rows)
    nnz = 0
    for i in rows:
        sp, ep = csr.row_extent(i)
        rl = ep - sp
        nnz += rl

    # allocate result arrays
    rowptrs = np.empty(nr + 1, dtype=np.int32)
    colinds = np.empty(nnz, dtype=np.int32)
    values = np.empty(nnz, dtype=csr.values.dtype)
    pos = 0

    # copy each item's rows to result array
    for ii, i in enumerate(rows):
        # get row start/end
        sp, ep = csr.row_extent(i)
        # how many tags for this item?
        itc = ep - sp
        end = pos + itc
        # copy values
        colinds[pos:end] = csr.colinds[sp:ep]
        values[pos:end] = csr.values[sp:ep]
        # set offset
        rowptrs[ii] = pos
        # update position for storing results
        pos = end
    rowptrs[nr] = pos

    return CSR(nr, csr.ncols, nnz, rowptrs, colinds, values)


@njit(nogil=True)
def sort_rows(csr):
    "Sort the rows of a CSR by increasing column index"
    for i in range(csr.nrows):
        sp, ep = csr.row_extent(i)
        # bubble-sort so it's super-fast on sorted arrays
        swapped = True
        while swapped:
            swapped = False
            for j in range(sp, ep - 1):
                if csr.colinds[j] > csr.colinds[j + 1]:
                    _util.maybe_swap(csr.colinds, j, j + 1)
                    _util.maybe_swap(csr.values, j, j + 1)
                    swapped = True


@njit(nogil=True)
def _transpose_values(csr):
    "Transpose a CSR with its values."
    brp = np.zeros(csr.ncols + 1, csr.rowptrs.dtype)
    bci = np.zeros(csr.nnz, np.int32)
    bvs = np.zeros(csr.nnz, np.float64)

    # count elements
    for i in range(csr.nrows):
        ars, are = csr.row_extent(i)
        for jj in range(ars, are):
            j = csr.colinds[jj]
            brp[j + 1] += 1

    # convert to pointers
    for j in range(csr.ncols):
        brp[j+1] = brp[j] + brp[j+1]

    # construct results
    for i in range(csr.nrows):
        ars, are = csr.row_extent(i)
        for jj in range(ars, are):
            j = csr.colinds[jj]
            bci[brp[j]] = i
            bvs[brp[j]] = csr.values[jj]
            brp[j] += 1

    # restore pointers
    for i in range(csr.ncols - 1, 0, -1):
        brp[i] = brp[i - 1]
    brp[0] = 0

    return create(csr.ncols, csr.nrows, csr.nnz, brp, bci, bvs)


@njit(nogil=True)
def _transpose_structure(csr):
    "Transpose a CSR, structure-only."
    brp = np.zeros(csr.ncols + 1, csr.rowptrs.dtype)
    bci = np.zeros(csr.nnz, np.int32)

    # count elements
    for i in range(csr.nrows):
        ars, are = csr.row_extent(i)
        for jj in range(ars, are):
            j = csr.colinds[jj]
            brp[j + 1] += 1

    # convert to pointers
    for j in range(csr.ncols):
        brp[j + 1] = brp[j] + brp[j+1]

    # construct results
    for i in range(csr.nrows):
        ars, are = csr.row_extent(i)
        for jj in range(ars, are):
            j = csr.colinds[jj]
            bci[brp[j]] = i
            brp[j] += 1

    # restore pointers
    for i in range(csr.ncols - 1, 0, -1):
        brp[i] = brp[i - 1]
    brp[0] = 0

    return create_novalues(csr.ncols, csr.nrows, csr.nnz, brp, bci)


def transpose(csr, include_values):
    if csr.values is None:
        include_values = False

    if include_values:
        return _transpose_values(csr)
    else:
        return _transpose_structure(csr)



import sys
from typing import Optional
import dataclasses
from numba.core.types.functions import _ResolutionFailures
import numpy as np
from numba import njit, config
from numba.extending import overload
from numba.core.types import StructRef, intc, float64
from numba.experimental import structref

from csr import CSR
from ._api import *  # noqa: F403
from csr.constructors import create_empty

__all__ = [
    'mkl_h',
    'to_handle',
    'from_handle',
    'release_handle'
]


@structref.register
class mkl_h_type(StructRef):
    "Internal Numba type for MKL handles"
    pass


if config.DISABLE_JIT:
    @dataclasses.dataclass
    class mkl_h:
        H: int
        nrows: int
        ncols: int
        csr_ref: Optional[np.ndarray]
else:
    class mkl_h(structref.StructRefProxy):
        """
        Type for MKL handles.  Opaque, do not use directly.
        """

    structref.define_proxy(mkl_h, mkl_h_type, ['H', 'nrows', 'ncols', 'csr_ref'])


def _make_handle_impl(csr):
    "Make a handle from a known-constructable CSR"
    _sp = ffi.from_buffer(csr.rowptrs)
    _cols = ffi.from_buffer(csr.colinds)
    vs = csr.values
    assert vs.size == csr.nnz
    _vals = ffi.from_buffer(vs)
    h = lk_mkl_spcreate(csr.nrows, csr.ncols, _sp, _cols, _vals)
    lk_mkl_spopt(h)
    return mkl_h(h, csr.nrows, csr.ncols, csr)


_make_handle = njit(_make_handle_impl)


def to_handle(csr: CSR) -> mkl_h:
    if csr.nnz == 0:
        # empty matrices don't really work
        return mkl_h(0, csr.nrows, csr.ncols, None)

    norm = csr._normalize(np.float64, np.intc)
    return _make_handle(norm)


@overload(to_handle)
def to_handle_jit(csr):
    if csr.ptr_type.dtype != intc:
        raise TypeError('MKL requires intc row pointers')

    if csr.has_values:
        vt = csr.val_type.dtype
    else:
        vt = None

    def mkh(csr):
        vs = csr._required_values().astype(np.float64)
        csr2 = CSR(csr.nrows, csr.ncols, csr.nnz, csr.rowptrs, csr.colinds, vs)

        if csr.nnz == 0:
            return mkl_h(0, csr.nrows, csr.ncols, csr2)

        return _make_handle(csr2)

    return mkh


@njit
def from_handle(h: mkl_h) -> CSR:
    if not h.H:
        return create_empty(h.nrows, h.ncols)

    rvp = lk_mkl_spexport_p(h.H)
    if rvp is None:
        return None

    nrows = lk_mkl_spe_nrows(rvp)
    ncols = lk_mkl_spe_ncols(rvp)

    sp = lk_mkl_spe_row_sp(rvp)
    ep = lk_mkl_spe_row_ep(rvp)
    cis = lk_mkl_spe_colinds(rvp)
    vs = lk_mkl_spe_values(rvp)

    rowptrs = np.zeros(nrows + 1, dtype=np.intc)
    nnz = 0
    for i in range(nrows):
        nnz += ep[i] - sp[i]
        rowptrs[i + 1] = nnz
    assert nnz == ep[nrows - 1]

    colinds = np.zeros(nnz, dtype=np.intc)
    values = np.zeros(nnz)

    for i in range(nrows):
        rs = rowptrs[i]
        re = rowptrs[i + 1]
        ss = sp[i]
        for j in range(re - rs):
            colinds[rs + j] = cis[ss + j]
            values[rs + j] = vs[ss + j]

    lk_mkl_spe_free(rvp)

    return CSR(nrows, ncols, nnz, rowptrs, colinds, values)


@njit
def order_columns(h):
    """
    Sort matrix rows in increasing column order.
    """
    if h.H:
        lk_mkl_sporder(h.H)


@njit
def release_handle(h: mkl_h):
    if h.H:
        lk_mkl_spfree(h.H)
    h.H = 0



from textwrap import dedent
from _pytest.python_api import approx
import numpy as np
from numpy.testing._private.utils import assert_equal
import scipy.sparse as sps

from csr import CSR
from csr.test_utils import csrs, matrices, mm_pairs, sparse_matrices

from numba import njit

from pytest import raises
from hypothesis import given, assume, settings, HealthCheck
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph


_getters = {}

for __a in ['ncols', 'nrows', 'nnz', 'rowptrs', 'colinds', 'values']:
    __env = {'njit': njit, 'defs': {}}
    exec(dedent(f'''
    @njit
    def _get_{__a}(csr):
        return csr.{__a}
    defs['getter'] = _get_{__a}
    '''), __env)
    _getters[__a] = __env['defs']['getter']


def _get_attr(csr, name):
    getter = _getters[name]
    return getter(csr)


@given(csrs())
def test_access_fields(csr):
    assert _get_attr(csr, 'nrows') == csr.nrows
    assert _get_attr(csr, 'ncols') == csr.ncols
    assert _get_attr(csr, 'nnz') == csr.nnz
    assert all(_get_attr(csr, 'rowptrs') == csr.rowptrs)
    assert all(_get_attr(csr, 'colinds') == csr.colinds)
    if csr.values is None:
        assert _get_attr(csr, 'values') is None
    else:
        assert all(_get_attr(csr, 'values') == csr.values)


@njit
def _extent(csr, row):
    return csr.row_extent(row)


@given(csrs())
def test_csr_row_extent(csr):

    for i in range(csr.nrows):
        sp, ep = _extent(csr, i)
        assert sp == csr.rowptrs[i]
        assert ep == csr.rowptrs[i + 1]


@njit
def _row(csr, row):
    return csr.row(row)


@given(csrs())
def test_csr_row(csr):

    for i in range(csr.nrows):
        cr = csr.row(i)
        nr = _row(csr, i)
        assert nr.shape == cr.shape
        assert all(nr == cr)


@njit
def _row_cvs(csr, row):
    cs = csr.row_cs(row)
    vs = csr.row_vs(row)
    return cs, vs


@given(csrs())
def test_csr_row_cvs(csr):
    for i in range(csr.nrows):
        ccs = csr.row_cs(i)
        cvs = csr.row_vs(i)
        ncs, nvs = _row_cvs(csr, i)
        assert ncs.shape == ccs.shape
        assert nvs.shape == cvs.shape
        assert all(ncs == ccs)
        assert all(nvs == cvs)


@njit
def _transpose(csr):
    return csr.transpose()


@given(csrs())
def test_transpose(csr):
    t = _transpose(csr)

    assert t.nrows == csr.ncols
    assert t.ncols == csr.nrows
    assert t.nnz == csr.nnz

    t2 = csr.transpose()
    assert all(t.rowptrs == t2.rowptrs)
    assert all(t.colinds == t2.colinds)

    if csr.values is not None:
        assert t.values is not None
        assert all(t.values == t2.values)
    else:
        assert t.values is None


@njit
def _transpose_s(csr):
    return csr.transpose_structure()


@given(csrs())
def test_transpose_structure(csr):
    t = _transpose_s(csr)

    assert t.nrows == csr.ncols
    assert t.ncols == csr.nrows
    assert t.nnz == csr.nnz

    t2 = csr.transpose()
    assert all(t.rowptrs == t2.rowptrs)
    assert all(t.colinds == t2.colinds)

    assert t.values is None


@njit
def _mult(A, B, transpose):
    return A.multiply(B, transpose)


@given(mm_pairs(), st.booleans())
def test_numba_mult(pair, transpose):
    A, B = pair
    C = A @ B

    A = CSR.from_scipy(A)
    B = CSR.from_scipy(B)

    if transpose:
        B = B.transpose()

    res = _mult(A, B, transpose)

    cnr, cnc = C.shape
    assert res.nrows == cnr
    assert res.ncols == cnc
    assert res.nnz == C.nnz


@njit
def _mult_vec(A, x):
    return A.mult_vec(x)


@given(st.data())
def test_numba_mult_vec(data):
    A = data.draw(csrs())
    vals = st.floats(-100, 100)
    x = data.draw(nph.arrays(np.float64, A.ncols, elements=vals))

    y = _mult_vec(A, x)

    assert y.shape == (A.nrows,)
    assert y == approx(A.to_scipy() @ x, nan_ok=True, rel=1.0e-5)



import logging

from hypothesis import settings, HealthCheck
from pytest import fixture, skip
from csr import CSR
from csr.kernels import use_kernel, get_kernel

# turn off Numba logging
logging.getLogger('numba').setLevel(logging.INFO)

KERNELS = ["scipy", "numba", "mkl"]
DISABLED_KERNELS = []
try:
    import csr.kernels.mkl  # noqa: F401
except ImportError:
    DISABLED_KERNELS.append("mkl")


# set up fixtures
@fixture(scope="module", params=KERNELS)
def kernel(request):
    """
    Fixture for variable CSR kernels.  This fixture is parameterized, so if you
    write a test function with a parameter ``kernel`` as its first parameter, it
    will be called once for each kernel under active test.
    """
    if request.param in DISABLED_KERNELS:
        skip(f'kernel {request.param} is disabled')

    with use_kernel(request.param):
        k = get_kernel()
        # warm-up the kernel
        m = CSR.empty(1, 1)
        h = k.to_handle(m)
        k.release_handle(h)
        del h, m
        yield k


# set up profiles
settings.register_profile('default', deadline=2500)
settings.register_profile('large', settings.get_profile('default'),
                          max_examples=5000, deadline=None)
settings.register_profile('fast', max_examples=50)
settings.register_profile('nojit', settings.get_profile('fast'),
                          deadline=None, suppress_health_check=HealthCheck.all())
settings.load_profile('default')



import logging
import numpy as np
import scipy.sparse as sps

from csr import CSR
from csr.test_utils import csrs, csr_slow, sparse_matrices

from pytest import mark, approx, raises
from hypothesis import given, assume, settings, HealthCheck
import hypothesis.strategies as st
import hypothesis.extra.numpy as nph

_log = logging.getLogger(__name__)


@csr_slow()
@given(st.data())
def test_subset_rows(data):
    nrows = data.draw(st.integers(5, 100))
    ncols = data.draw(st.integers(1, 100))
    dens = data.draw(st.floats(0, 1))
    beg = data.draw(st.integers(0, nrows - 1))
    end = data.draw(st.integers(beg, nrows - 1))

    spm = sps.random(nrows, ncols, dens, format='csr')
    csr = CSR.from_scipy(spm)

    m2 = csr.subset_rows(beg, end)
    assert m2.nrows == end - beg

    for i in range(m2.nrows):
        assert all(m2.row_cs(i) == csr.row_cs(beg+i))
        assert all(m2.row_vs(i) == csr.row_vs(beg+i))


@csr_slow()
@given(st.data())
def test_pick_rows(data):
    include = data.draw(st.booleans())
    csr = data.draw(csrs())
    nrows = csr.nrows

    to_pick = st.integers(0, nrows * 10)
    r_range = st.integers(0, nrows - 1)
    rows = data.draw(nph.arrays(np.int32, to_pick, elements=r_range))

    _log.debug('picking %d rows from %d-row matrix', len(rows), nrows)

    sub = csr.pick_rows(rows, include_values=include)
    assert sub.nrows == len(rows)
    assert len(sub.rowptrs) == sub.nrows + 1
    if include and csr.values is not None:
        assert sub.values is not None
    else:
        assert sub.values is None

    for i, r in enumerate(rows):
        osp, oep = csr.row_extent(r)
        sp, ep = sub.row_extent(i)
        assert oep - osp == ep - sp

        assert all(csr.row_cs(r) == sub.row_cs(i))
        if sub.values is not None:
            assert all(csr.row_vs(r) == sub.row_vs(i))


@given(csrs())
def test_sort_rows(csr):
    tv = np.ones(csr.ncols)
    x1 = csr.mult_vec(tv)
    csr.sort_rows()
    assert all(all(np.diff(csr.row_cs(i)) > 0) for i in range(csr.nrows))
    x2 = csr.mult_vec(tv)
    assert x2 == approx(x1)


@given(csrs())
def test_kernel_sort_rows(kernel, csr):
    tv = np.ones(csr.ncols)
    x1 = csr.mult_vec(tv)
    h = kernel.to_handle(csr.R)
    kernel.order_columns(h)
    c2 = kernel.from_handle(h)
    kernel.release_handle(h)
    assert all(all(np.diff(c2.row_cs(i)) > 0) for i in range(csr.nrows))
    x2 = c2.mult_vec(tv)
    assert x2 == approx(x1)


@csr_slow()
@given(sparse_matrices())
def test_mean_center(spm):
    # assume(spm.nnz >= 10)
    csr = CSR.from_scipy(spm)

    m2 = csr.normalize_rows('center')
    assert len(m2) == csr.nrows
    rnnz = csr.row_nnzs()

    for i in range(csr.nrows):
        vs = csr.row_vs(i)
        spr = spm[i, :].toarray()
        if rnnz[i] > 0:
            assert m2[i] == approx(np.sum(spr) / rnnz[i])
            assert np.mean(vs) == approx(0.0)
            assert vs + m2[i] == approx(spr[0, csr.row_cs(i)])


@csr_slow()
@given(sparse_matrices())
def test_unit_norm(spm):
    # assume(spm.nnz >= 10)
    csr = CSR.from_scipy(spm)

    m2 = csr.normalize_rows('unit')
    assert len(m2) == csr.nrows

    for i in range(csr.nrows):
        vs = csr.row_vs(i)
        if len(vs) > 0:
            assert np.linalg.norm(vs) == approx(1.0)
            assert vs * m2[i] == approx(spm.getrow(i).toarray()[0, csr.row_cs(i)])


@csr_slow()
@given(sparse_matrices())
def test_filter(mat):
    assume(mat.nnz > 0)
    assume(not np.all(mat.data <= 0))  # we have to have at least one to retain
    csr = CSR.from_scipy(mat)
    csrf = csr.filter_nnzs(csr.values > 0)
    assert all(csrf.values > 0)
    assert csrf.nnz <= csr.nnz

    for i in range(csr.nrows):
        spo, epo = csr.row_extent(i)
        spf, epf = csrf.row_extent(i)
        assert epf - spf <= epo - spo

    d1 = csr.to_scipy().toarray()
    df = csrf.to_scipy().toarray()
    d1[d1 < 0] = 0
    assert df == approx(d1)


# @csr_slow()
@given(sparse_matrices())
def test_shard(mat):
    SHARD_SIZE = 1000
    csr = CSR.from_scipy(mat)

    shards = csr._shard_rows(SHARD_SIZE)
    # we have the whole matrix
    assert sum(s.nnz for s in shards) == csr.nnz
    # everything is in spec
    assert all(s.nnz <= SHARD_SIZE for s in shards)

    # all row counts match
    assert np.all(np.concatenate([s.row_nnzs() for s in shards]) == csr.row_nnzs())
    # all column indices match
    assert np.all(np.concatenate([s.colinds for s in shards]) == csr.colinds)
    # all values match
    assert np.all(np.concatenate([s.values for s in shards]) == csr.values)

    # we can reassemble the shards
    csr2 = CSR._assemble_shards(shards)
    assert csr2.nrows == csr.nrows
    assert csr2.ncols == csr.ncols
    assert csr2.nnz == csr.nnz
    assert np.all(csr2.rowptrs == csr.rowptrs)
    assert np.all(csr2.colinds == csr.colinds)
    assert np.all(csr2.values == csr.values)



"""
Numba-accessible constructors for CSRs.
"""

from .csr import CSR

import numpy as np
from numba import njit


@njit
def create_empty(nrows, ncols):
    """
    Create an empty CSR of the specified size.

    .. note:: This function can be used from Numba.
    """
    nrows = np.int32(nrows)
    ncols = np.int32(ncols)
    rowptrs = np.zeros(nrows + 1, dtype=np.intc)
    colinds = np.zeros(0, dtype=np.intc)
    values = np.zeros(0)
    return CSR(np.int32(nrows), np.int32(ncols), 0, rowptrs, colinds, values)


@njit
def create_novalues(nrows, ncols, nnz, rowptrs, colinds):
    """
    Create a CSR without values.
    """
    nrows = np.int32(nrows)
    ncols = np.int32(ncols)
    return CSR(nrows, ncols, nnz, rowptrs, colinds, None)


@njit
def create(nrows, ncols, nnz, rowptrs, colinds, values):
    """
    Create a CSR.
    """
    nrows = np.int32(nrows)
    ncols = np.int32(ncols)
    return CSR(nrows, ncols, nnz, rowptrs, colinds, values)


@njit
def create_from_sizes(nrows, ncols, sizes):
    """
    Create a CSR with uninitialized values and specified row sizes.

    This function is Numba-accessible, but is limited to creating matrices with fewer
    than :math:`2^{31}` nonzero entries and present value arrays.

    Args:
        nrows(int): the number of rows
        ncols(int): the number of columns
        sizes(numpy.ndarray): the number of nonzero values in each row
    """
    nrows = np.int32(nrows)
    ncols = np.int32(ncols)
    nnz = np.sum(sizes)
    assert nnz >= 0
    rowptrs = np.zeros(nrows + 1, dtype=np.int32)
    for i in range(nrows):
        rowptrs[i + 1] = rowptrs[i] + sizes[i]
    colinds = np.full(nnz, -1, dtype=np.intc)
    values = np.full(nnz, np.nan)
    return CSR(nrows, ncols, nnz, rowptrs, colinds, values)



"""
Python API for CSR matrices.
"""

import warnings
import logging
import numpy as np
import scipy.sparse as sps

from numba import config
from numba.experimental import structref

from csr.kernels import get_kernel, releasing
from . import _struct, _rows

INTC = np.iinfo(np.intc)
_log = logging.getLogger(__name__)

# ugly hack for a bug on Numba < 0.53
if config.DISABLE_JIT:
    class _csr_base:
        def __init__(self, nrows, ncols, nnz, ptrs, inds, vals, _cast=True):
            self.nrows = nrows
            self.ncols = ncols
            self.nnz = nnz
            if _cast and np.max(ptrs, initial=0) <= INTC.max:
                self.rowptrs = np.require(ptrs, np.intc, 'C')
            else:
                self.rowptrs = np.require(ptrs, requirements='C')
            self.colinds = np.require(inds, np.intc, 'C')
            if vals is not None:
                self._values = np.require(vals, requirements='C')
            else:
                self._values = None

        def _numba_box_(self, *args):
            raise NotImplementedError()

    NUMBA_ENABLED = False

else:
    _csr_base = structref.StructRefProxy
    NUMBA_ENABLED = True


class CSR(_csr_base):
    """
    Simple compressed sparse row matrix.  This is like :py:class:`scipy.sparse.csr_matrix`, with
    a few useful differences:

    * The value array is optional, for cases in which only the matrix structure is required.
    * The value array, if present, is always double-precision.
    * It is usable from code compiled in Numba's nopython mode.

    You generally don't want to create this class yourself with the constructor.  Instead, use one
    of its class or static methods.  If you do use the constructor, be advised that the class may
    reuse the arrays that you pass, but does not guarantee that they will be used.

    Not all methods are available from Numba, and a few have restricted signatures.  The
    documentation for each method notes deviations when in Numba-compiled code.

    At the Numba level, matrices with and without value arrays have different types. For the
    most part, this is transparent, but if you want to write a Numba function that works on
    the values array but only if it is present, it requires writing two versions of the
    function and using :py:func:`numba.extending.overload` to dispatch to the correct one.
    There are several examples of doing this in the CSR source code. The method
    :py:meth:`CSRType.has_values` lets you quickly see if a CSR type instance has
    values or not.

    Attributes:
        nrows(int): the number of rows.
        ncols(int): the number of columns.
        nnz(int): the number of entries.
        rowptrs(numpy.ndarray): the row pointers.
        colinds(numpy.ndarray): the column indices.
        values(numpy.ndarray or None): the values.
    """

    def __new__(cls, nrows, ncols, nnz, rps, cis, vs, _cast=True):
        assert nrows >= 0
        assert nrows <= INTC.max
        assert ncols >= 0
        assert ncols <= INTC.max
        assert nnz >= 0
        nrows = np.intc(nrows)
        ncols = np.intc(ncols)

        if _cast:
            cis = np.require(cis, np.intc, 'C')
            if nnz <= INTC.max:
                rps = np.require(rps, np.intc, 'C')
            else:
                rps = np.require(rps, np.int64, 'C')
            if vs is not None:
                vs = np.require(vs, requirements='C')

        if NUMBA_ENABLED:
            return _csr_base.__new__(cls, nrows, ncols, nnz, rps, cis, vs)
        else:
            return _csr_base.__new__(cls)

    @classmethod
    def empty(cls, nrows, ncols, row_nnzs=None, values=True):
        """
        Create an uninitialized CSR matrix.

        Args:
            nrows(int): the number of rows.
            ncols(int): the number of columns.
            row_nnzs(array-like):
                the number of nonzero entries for each row, or None for an empty matrix.
            values(bool, str, or numpy.dtype):
                whether it has values or only structure; can be a NumPy data type to
                specify a type other than `f8`.
        """
        from .constructors import create_empty
        assert nrows >= 0
        assert ncols >= 0
        if row_nnzs is not None:
            assert len(row_nnzs) == nrows
            nnz = np.sum(row_nnzs, dtype=np.int64)
            assert nnz >= 0
            rp_dtype = np.intc if nnz <= INTC.max else np.int64
            rps = np.zeros(nrows + 1, dtype=rp_dtype)
            np.cumsum(row_nnzs, dtype=rp_dtype, out=rps[1:])
            cis = np.zeros(nnz, dtype=np.int32)

            if values is True:
                vs = np.zeros(nnz)
            elif values:
                vs = np.zeros(nnz, dtype=values)
            else:
                vs = None

            return cls(nrows, ncols, nnz, rps, cis, vs)
        else:
            return create_empty(nrows, ncols)

    @classmethod
    def from_coo(cls, rows, cols, vals, shape=None, *, rpdtype=np.intc):
        """
        Create a CSR matrix from data in COO format.

        Args:
            rows(array-like): the row indices.
            cols(array-like): the column indices.
            vals(array-like): the data values; can be ``None``.
            shape(tuple): the array shape, or ``None`` to infer from row & column indices.
        """
        from .structure import from_coo
        if shape is not None:
            nrows, ncols = shape
            assert np.max(rows, initial=0) < nrows
            assert np.max(cols, initial=0) < ncols
        else:
            nrows = np.max(rows) + 1
            ncols = np.max(cols) + 1

        nnz = len(rows)
        assert len(cols) == nnz
        assert vals is None or len(vals) == nnz

        rowptrs, cols, vals = from_coo(nrows, rows, cols, vals)
        return cls(nrows, ncols, nnz, rowptrs, cols, vals)

    @classmethod
    def from_scipy(cls, mat, copy=True):
        """
        Convert a scipy sparse matrix to a CSR.

        Args:
            mat(scipy.sparse.spmatrix): a SciPy sparse matrix.
            copy(bool): if ``False``, reuse the SciPy storage if possible.

        Returns:
            CSR: a CSR matrix.
        """
        if not sps.isspmatrix_csr(mat):
            mat = mat.tocsr(copy=copy)
        rp = np.require(mat.indptr, np.intc, 'C')
        if copy and rp is mat.indptr:
            rp = rp.copy()
        cs = np.require(mat.indices, np.intc, 'C')
        if copy and cs is mat.indices:
            cs = cs.copy()
        vs = mat.data.copy() if copy else mat.data
        return cls(mat.shape[0], mat.shape[1], mat.nnz, rp, cs, vs)

    def to_scipy(self):
        """
        Convert a CSR matrix to a SciPy :py:class:`scipy.sparse.csr_matrix`.  Avoids copying
        if possible.

        Args:
            self(CSR): A CSR matrix.

        Returns:
            scipy.sparse.csr_matrix:
                A SciPy sparse matrix with the same data.
        """
        values = self.values
        if values is None:
            values = np.full(self.nnz, 1.0)
        return sps.csr_matrix((values, self.colinds, self.rowptrs), shape=(self.nrows, self.ncols))

    if _csr_base is structref.StructRefProxy:
        nrows = property(_struct.get_nrows)
        ncols = property(_struct.get_ncols)
        nnz = property(_struct.get_nnz)
        rowptrs = property(_struct.get_rowptrs)
        colinds = property(_struct.get_colinds)

    @property
    def values(self):
        if NUMBA_ENABLED:
            return _struct.get_values(self)
        else:
            return self._values

    @values.setter
    def values(self, vs: np.ndarray):
        if vs is None:
            new = CSR(self.nrows, self.ncols, self.nnz, self.rowptrs, self.colinds, None)
        else:
            if len(vs) < self.nnz:
                raise ValueError('value array too small')
            elif len(vs) > self.nnz:
                vs = vs[:self.nnz]

            new = CSR(self.nrows, self.ncols, self.nnz, self.rowptrs, self.colinds, vs)

        if NUMBA_ENABLED:
            # replace our internals
            self._type = new._type
            self._meminfo = new._meminfo
        else:
            self._values = new._values

    def _required_values(self):
        """
        Get the value array, returning an array of 1s if it is not present.
        """
        vs = self.values
        if vs is None:
            return np.ones(self.nnz)
        else:
            return vs

    def _e_value(self, i):
        """
        Get the value of a particular element, returning 1 if values is undefined.
        """
        vs = self.values
        if vs is not None:
            return vs[i]
        else:
            return 1.0

    def _normalize(self, val_dtype=np.float64, ptr_dtype=None):
        """
        Normalize the matrix into a predictable structure and type.  It avoids copying
        if possible.

        .. note:: This method is not available from Numba.

        Args:
            val_dtype(np.dtype or None or boolean):
                The value data type.  If ``False``, drop the value array.  If ``None``,
                leave unchanged.
            ptr_dtype(np.dtype or None):
                The row pointer data type.  If ``None``, leave rows untransformed.
        Returns:
            CSR: the transformed CSR matrix.
        """

        if ptr_dtype:
            info = np.iinfo(ptr_dtype)
            if self.nnz > info.max:
                raise ValueError(f'type {ptr_dtype} cannot address {self.nnz} entries')
            rps = np.require(self.rowptrs, ptr_dtype)
        else:
            rps = self.rowptrs

        if val_dtype:
            if self.values is None:
                vs = np.ones(self.nnz, val_dtype)
            else:
                vs = np.require(self.values, val_dtype)
        elif val_dtype is False:
            vs = None
        else:
            vs = self.values

        return CSR(self.nrows, self.ncols, self.nnz, rps, self.colinds, vs, _cast=False)

    @property
    def R(self):
        warnings.warn('.R deprecated, use CSR directly', DeprecationWarning)
        return self

    def copy(self, include_values=True, *, copy_structure=True):
        """
        Create a copy of this CSR.

        Args:
            include_values(bool): whether to copy the values or only the structure.
            copy_structure(bool):
                whether to copy the structure (index & pointers) or share with the original matrix.
        """
        values = self.values
        if include_values and values is not None:
            values = np.copy(values)
        else:
            values = None
        rps = self.rowptrs
        cis = self.colinds
        if copy_structure:
            rps = np.copy(rps)
            cis = np.copy(cis)
        return CSR(self.nrows, self.ncols, self.nnz,
                   rps, cis, values)

    def sort_rows(self):
        """
        Sort the rows of this matrix in column order.  This is an **in-place operation**.
        """
        from .structure import sort_rows
        sort_rows(self)

    def subset_rows(self, begin, end):
        """
        Subset the rows in this matrix.

        .. note:: This method is not available from Numba.

        Args:
            begin(int): the first row index to include.
            end(int): one past the last row to include.

        Returns:
            CSR: the matrix only containing a subset of the rows.  It shares storage
                with the original matrix to the extent possible.
        """
        from .structure import subset_rows
        return subset_rows(self, begin, end)

    def pick_rows(self, rows, *, include_values=True):
        """
        Pick rows from this matrix.  A row may appear more than once.

        .. note:: This method is not available from Numba.

        Args:
            rows(numpy.ndarray): the row indices to select.
            include_values(bool): whether to include values if present

        Returns:
            CSR: the matrix containing the specified rows.
        """
        from .structure import _pick_rows, _pick_rows_nvs
        if include_values and self.values is not None:
            return _pick_rows(self, rows)
        else:
            return _pick_rows_nvs(self, rows)

    def rowinds(self) -> np.ndarray:
        """
        Get the row indices from this array.  Combined with :py:attr:`colinds` and
        :py:attr:`values`, this can form a COO-format sparse matrix.
        """
        return _rows.all_indices(self)

    def row(self, row):
        """
        Return a row of this matrix as a dense ndarray.

        Args:
            row(int): the row index.

        Returns:
            numpy.ndarray:
                the row, with 0s in the place of missing values.  If the CSR only
                stores matrix structure, the returned vector has 1s where the CSR
                records an entry.
        """
        return _rows.array(self, row)

    def row_extent(self, row):
        """
        Get the extent of a row in the underlying column index and value arrays.

        Args:
            row(int): the row index.

        Returns:
            tuple: ``(s, e)``, where the row occupies positions :math:`[s, e)` in the
            CSR data.
        """
        return _rows.extent(self, row)

    def row_cs(self, row):
        """
        Get the column indcies for the stored values of a row.
        """
        return _rows.cs(self, row)

    def row_vs(self, row):
        """
        Get the stored values of a row.  If only the matrix structure is stored, this
        returns a vector of 1s.
        """
        return _rows.vs(self, row)

    def row_nnzs(self):
        """
        Get a vector of the number of nonzero entries in each row.

        .. note:: This method is not available from Numba.

        Returns:
            numpy.ndarray: the number of nonzero entries in each row.
        """
        return np.diff(self.rowptrs)

    def normalize_rows(self, normalization):
        """
        Normalize the rows of the matrix.

        .. note:: The normalization *ignores* missing values instead of treating
                  them as 0.

        .. note:: This method is not available from Numba.

        Args:
            normalization(str):
                The normalization to perform. Can be one of:

                * ``'center'`` - center rows about the mean
                * ``'unit'`` - convert rows to a unit vector

        Returns:
            numpy.ndarray:
                The normalization values for each row.
        """
        from . import transform
        if normalization == 'center':
            return transform.center_rows(self)
        elif normalization == 'unit':
            return transform.unit_rows(self)
        else:
            raise ValueError('unknown normalization: ' + normalization)

    def transpose(self, include_values=True):
        """
        Transpose a CSR matrix.

        .. note::
            In Numba, this method takes no paramters.
            Call :py:meth:`transpose_structure` for a structure-only transpose.

        Args:
            include_values(bool): whether to include the values in the transpose.

        Returns:
            CSR: the transpose of this matrix (or, equivalently, this matrix in CSC format).
        """
        from .structure import transpose
        return transpose(self, include_values)

    def transpose_structure(self):
        """
        Tranpose the structure of a CSR matrix.  The resulting matrix has no values.
        """
        return self.transpose(False)

    def filter_nnzs(self, filt):
        """
        Filter the values along the full NNZ axis.

        .. note:: This method is not available from Numba.

        Args:
            filt(ndarray):
                a logical array of length :attr:`nnz` that indicates the values to keep.

        Returns:
            CSR: The filtered sparse matrix.
        """
        if len(filt) != self.nnz:
            raise ValueError('filter has length %d, expected %d' % (len(filt), self.nnz))
        rps2 = np.zeros_like(self.rowptrs)
        for i in range(self.nrows):
            sp, ep = self.row_extent(i)
            rlen = np.sum(filt[sp:ep])
            rps2[i + 1] = rps2[i] + rlen

        nnz2 = rps2[-1]
        assert nnz2 == np.sum(filt)

        cis2 = self.colinds[filt]
        vs = self.values
        vs2 = None if vs is None else vs[filt]

        return CSR(self.nrows, self.ncols, nnz2, rps2, cis2, vs2)

    def multiply(self, other, transpose=False):
        """
        Multiply this matrix by another.

        .. note:: In Numba, ``transpose`` is a mandatory positional argument.  Numba users
                  may wish to directly use the kernel API.

        Args:
            other(CSR): the other matrix.
            transpose(bool): if ``True``, compute :math:`AB^{T}` instead of :math:`AB`.

        Returns
            CSR: the product of the two matrices.
        """
        if transpose:
            assert self.ncols == other.ncols
        else:
            assert self.ncols == other.nrows

        K = get_kernel()

        # Helper for handling sharding
        def mul(A, b_h):
            with releasing(K.to_handle(A), K) as a_h:
                if transpose:
                    c_h = K.mult_abt(a_h, b_h)
                else:
                    c_h = K.mult_ab(a_h, b_h)
                with releasing(c_h, K):
                    crepr = K.from_handle(c_h)

            return crepr

        if self.nnz <= K.max_nnz:
            # Common / fast path - one matrix
            with releasing(K.to_handle(other), K) as b_h:
                return mul(self, b_h)
        else:
            # Too large, let's go sharding
            shards = self._shard_rows(K.max_nnz)
            with releasing(K.to_handle(other), K) as b_h:
                sparts = [mul(s, b_h) for s in shards]
            return CSR._assemble_shards(sparts)

    def mult_vec(self, v):
        """
        Multiply this matrix by a vector.

        Args:
            other(numpy.ndarray): A vector, of length `ncols`.

        Returns:
            numpy.ndarray: :math:`A\\vec{x}`, as a vector.
        """
        assert v.shape == (self.ncols,)
        K = get_kernel()
        if self.nnz <= K.max_nnz:
            with releasing(K.to_handle(self), K) as h:
                return K.mult_vec(h, v)
        else:
            shards = self._shard_rows(K.max_nnz)
            svs = []
            for s in shards:
                with releasing(K.to_handle(s), K) as h:
                    svs.append(K.mult_vec(h, v))
            return np.concatenate(svs)

    def _shard_rows(self, tgt_nnz):
        """
        Shard a matrix by rows to fit in a target size.
        """
        assert tgt_nnz > 0

        rest = self
        shards = []
        while rest.nnz > tgt_nnz:
            # find the first split point
            split = np.searchsorted(rest.rowptrs, tgt_nnz)
            # if the start of the found row is too large, back up by one
            if rest.rowptrs[split] > tgt_nnz:
                if split <= 1:
                    raise ValueError("row too large to fit in target matrix size")
                split -= 1

            _log.debug('splitting %s at %d (rp@s: %d)', rest, split, rest.rowptrs[split])
            shards.append(rest.subset_rows(0, split))
            rest = rest.subset_rows(split, rest.nrows)

        shards.append(rest)
        return shards

    @classmethod
    def _assemble_shards(cls, shards):
        """
        Reassemble a matrix from sharded rows.
        """
        nrows = sum(s.nrows for s in shards)
        ncols = max(s.ncols for s in shards)
        nnz = sum(s.nnz for s in shards)

        rps = np.zeros(nrows + 1, np.int64)
        rs = 0
        for s in shards:
            off = rps[rs]
            re = rs + s.nrows + 1
            rps[rs:re] = s.rowptrs + off
            rs += s.nrows

        assert rps[nrows] == nnz, f'{rps[nrows]} != {nnz}'

        cis = np.concatenate([s.colinds for s in shards])
        assert len(cis) == nnz
        if shards[0].values is not None:
            vs = np.concatenate([s.values for s in shards])
            assert len(vs) == nnz
        else:
            vs = None

        return cls(nrows, ncols, nnz, rps, cis, vs)

    def drop_values(self):
        """
        Remove the value array from this CSR.  This is an **in-place** operation.

        .. warning:: This method is deprecated.

        .. note:: This method is not available from Numba.
        """
        warnings.warn('drop_values is deprecated', DeprecationWarning)
        self.values = None

    def fill_values(self, value):
        """
        Fill the values of this CSR with the specified value.  If the CSR is
        structure-only, a value array is added.  This is an **in-place** operation.

        .. warning:: This method is deprecated.

        .. note:: This method is not available from Numba.
        """
        if self.values is not None:
            self.values[:] = value
        else:
            self.values = np.full(self.nnz, value, dtype='float64')

    def __str__(self):
        return '<CSR {}x{} ({} nnz)>'.format(self.nrows, self.ncols, self.nnz)

    def __repr__(self):
        repr = '<CSR {}x{} ({} nnz)'.format(self.nrows, self.ncols, self.nnz)
        repr += ' {\n'
        repr += '  rowptrs={}\n'.format(self.rowptrs)
        repr += '  colinds={}\n'.format(self.colinds)
        repr += '  values={}\n'.format(self.values)
        repr += '}>'
        return repr

    def __reduce__(self):
        args = (self.nrows, self.ncols, self.nnz, self.rowptrs, self.colinds, self.values, False)
        return (CSR, args)



"""
Various MKL-specific tests.
"""

import logging
from contextlib import contextmanager
from numba import njit, prange
import numpy as np

from csr import CSR
from csr.test_utils import csrs, has_memory

from pytest import skip, mark
from hypothesis import given

import test_multiply as tmm
import test_mult_vec as tmv


try:
    from csr.kernels import mkl
except ImportError:
    pytestmark = mark.skip("MKL is not available")

_log = logging.getLogger(__name__)


@contextmanager
def mkl_lim(lim=1000):
    "Limit MKL to a capacity of X"
    save = mkl.max_nnz
    try:
        mkl.max_nnz = lim
        yield lim
    finally:
        mkl.max_nnz = save
        pass


@njit
def make_handle(csr):
    return mkl.to_handle(csr)


@njit
def unhandle(h):
    return mkl.from_handle(h)


@given(csrs())
def test_csr_handle(csr):
    h = make_handle(csr)
    csr2 = unhandle(h)

    assert csr2.nrows == csr.nrows
    assert csr2.ncols == csr.ncols
    assert csr2.nnz == csr.nnz
    assert np.all(csr2.rowptrs == csr.rowptrs)
    assert np.all(csr2.colinds == csr.colinds)
    if csr.values is not None:
        assert np.all(csr2.values == csr.values)
    else:
        assert np.all(csr2.values == 1.0)


@njit(parallel=True)
def fill_rows(values, colinds, nrows, ncols, dense):
    for i in prange(nrows):
        s = i * dense
        e = s + dense
        values[s:e] = np.random.randn(e - s)
        colinds[s:e] = np.random.choice(ncols, dense, replace=False)


def test_mult_vec_lim():
    "Test matrix-vector multiply with limited kernel capacity"
    with mkl_lim():
        tmv.test_mult_vec(mkl)


def test_multiply_lim():
    "Test matrix-matrix multiply with limited kernel capacity"
    with mkl_lim():
        tmm.test_multiply(mkl)


def test_multiply_transpose_lim():
    "Select matrix-matrix transpose multiply with limited kernel capacity"
    with mkl_lim():
        tmm.test_multiply_transpose(mkl)


@mark.skipif(not has_memory(32), reason='insufficient memory')
def test_large_mult_vec():
    # 10M * 500 = 2.5B >= INT_MAX
    nrows = 10000000
    ncols = 500
    dense = 250
    nnz = nrows * dense

    rowptrs = np.arange(0, nnz + 1, dense, dtype=np.int64)

    assert len(rowptrs) == nrows + 1
    assert rowptrs[-1] == nnz

    try:
        _log.info('allocating indexes')
        colinds = np.empty(nnz, dtype=np.intc)
        _log.info('allocating values')
        values = np.zeros(nnz)
    except MemoryError:
        skip('insufficient memory')

    _log.info('randomizing array contents')
    fill_rows(values, colinds, nrows, ncols, dense)

    csr = CSR(nrows, ncols, nnz, rowptrs, colinds, values)

    v = np.random.randn(ncols)

    res = csr.mult_vec(v)

    assert res.shape == (nrows,)
    assert np.all(~np.isnan(res))



"""
Wire together the Numba and Python types.
"""

import numpy as np
from numba.extending import overload_method
from numba.experimental import structref

from ._struct import CSRType
from .csr import CSR
from . import _rows, structure

structref.define_proxy(CSR, CSRType, [
    'nrows', 'ncols', 'nnz',
    'rowptrs', 'colinds', 'values'
])


@overload_method(CSRType, 'row_extent')
def _csr_row_extent(csr, row):
    return _rows.extent


@overload_method(CSRType, 'row')
def _csr_row(csr, row):
    if csr.has_values:
        return _rows._array_vals
    else:
        return _rows._array_ones


@overload_method(CSRType, 'row_cs')
def _csr_row_cs(csr, row):
    return _rows.cs


@overload_method(CSRType, 'row_vs')
def _csr_row_vs(csr, row):
    if csr.has_values:
        return _rows._vs_vals
    else:
        return _rows._vs_ones


@overload_method(CSRType, 'rowinds')
def _csr_rowinds(csr, row):
    return _rows.all_indices


@overload_method(CSRType, 'transpose')
def _csr_transpose(csr):
    if csr.has_values:
        return lambda csr: structure._transpose_values(csr)
    else:
        return lambda csr: structure._transpose_structure(csr)


@overload_method(CSRType, 'transpose_structure')
def _csr_transpose_structure(csr):
    return lambda csr: structure._transpose_structure(csr)


@overload_method(CSRType, '_e_value')
def _csr_e_value(csr, i):
    def one(csr, i):
        return 1

    def val(csr, i):
        return csr.values[i]

    if csr.has_values:
        return val
    else:
        return one


@overload_method(CSRType, '_required_values')
def _csr_required_values(csr):
    def ones(csr):
        return np.ones(csr.nnz)

    def vals(csr):
        return csr.values

    if csr.has_values:
        return vals
    else:
        return ones


@overload_method(CSRType, 'multiply')
def _csr_multiply(csr, other, transpose):
    from . import kernel

    def mult(csr, other, transpose):
        ah = kernel.to_handle(csr)
        bh = kernel.to_handle(other)
        if transpose:
            ch = kernel.mult_abt(ah, bh)
        else:
            ch = kernel.mult_ab(ah, bh)

        kernel.release_handle(bh)
        kernel.release_handle(ah)

        result = kernel.from_handle(ch)
        kernel.release_handle(ch)

        return result

    return mult


@overload_method(CSRType, 'mult_vec')
def _csr_mult_vec(csr, x):
    from . import kernel

    def m_ax(csr, x):
        ah = kernel.to_handle(csr)
        y = kernel.mult_vec(ah, x)
        kernel.release_handle(ah)

        return y

    return m_ax



"""
Implementations of CSR row access functions.
"""

import numpy as np


def extent(csr, row):
    "Get the extent of a row in the matrix storage."
    sp = csr.rowptrs[row]
    ep = csr.rowptrs[row + 1]
    return sp, ep


def _array_vals(csr, row):
    "Get a row as a dense vector."
    v = np.zeros(csr.ncols)
    if csr.nnz == 0:
        return v

    sp, ep = csr.row_extent(row)
    cols = csr.colinds[sp:ep]
    v[cols] = csr.values[sp:ep]

    return v


def _array_ones(csr, row):
    v = np.zeros(csr.ncols)
    if csr.nnz == 0:
        return v

    sp, ep = csr.row_extent(row)
    cols = csr.colinds[sp:ep]
    v[cols] = 1

    return v


def array(csr, row):
    if csr.values is not None:
        return _array_vals(csr, row)
    else:
        return _array_ones(csr, row)


def cs(csr, row):
    "Get the column indices for a row."
    sp, ep = csr.row_extent(row)
    return csr.colinds[sp:ep]


def _vs_vals(csr, row):
    sp, ep = csr.row_extent(row)
    return csr.values[sp:ep]


def _vs_ones(csr, row):
    sp, ep = csr.row_extent(row)
    return np.full(ep - sp, 1.0)


def vs(csr, row):
    "Get the nonzero values for a row."
    if csr.values is not None:
        return _vs_vals(csr, row)
    else:
        return _vs_ones(csr, row)


def all_indices(csr):
    "Get the row indices for the nonzero values in a matrix."
    ris = np.zeros(csr.nnz, np.intc)
    for i in range(csr.nrows):
        sp, ep = csr.row_extent(i)
        ris[sp:ep] = i
    return ris



"""
Base definitions of CSR structure.
"""

from numba import njit
from numba.core import types
from numba.experimental import structref


@structref.register
class CSRType(types.StructRef):
    def preprocess_fields(self, fields):
        return tuple((name, types.unliteral(typ)) for name, typ in fields)

    @property
    def has_values(self):
        "Query whether this CSR type has values."
        return not isinstance(self.field_dict['values'], types.NoneType)

    @property
    def ptr_type(self):
        "The type of pointers into the data."
        return self.field_dict['rowptrs']

    @property
    def val_type(self):
        "The value type for the CSR"
        return self.field_dict['values']


@njit
def get_nrows(self):
    return self.nrows


@njit
def get_ncols(self):
    return self.ncols


@njit
def get_nnz(self):
    return self.nnz


@njit
def get_rowptrs(self):
    return self.rowptrs


@njit
def get_colinds(self):
    return self.colinds


@njit
def get_values(self):
    return self.values



"""
Data transformation implementations.
"""

import logging
import numpy as np
from numba import njit

_log = logging.getLogger(__name__)


@njit(nogil=True)
def center_rows(csr):
    "Mean-center the nonzero values of each row of a CSR."
    means = np.zeros(csr.nrows)
    for i in range(csr.nrows):
        sp, ep = csr.row_extent(i)
        if sp == ep:
            continue  # empty row
        vs = csr.row_vs(i)
        m = np.mean(vs)
        means[i] = m
        csr.values[sp:ep] -= m

    return means


@njit(nogil=True)
def unit_rows(csr):
    "Normalize the rows of a CSR to unit vectors."
    norms = np.zeros(csr.nrows)
    for i in range(csr.nrows):
        sp, ep = csr.row_extent(i)
        if sp == ep:
            continue  # empty row
        vs = csr.row_vs(i)
        m = np.linalg.norm(vs)
        norms[i] = m
        csr.values[sp:ep] /= m

    return norms



"""
Tests for creating kernel handles.
"""

from csr.test_utils import csrs

from hypothesis import given


@given(csrs())
def test_make_handle(kernel, csr):
    h = kernel.to_handle(csr.R)
    try:
        assert h is not None
        c2 = kernel.from_handle(h)
        assert c2.nrows == csr.nrows
        assert c2.ncols == csr.ncols
        assert c2.nnz == csr.nnz
    finally:
        kernel.release_handle(h)



"""
Compressed Sparse Row matrices for Python, with Numba API.
"""

__version__ = "0.4.1"
__all__ = [
    'CSR'
]

from .csr import CSR

# set up the Numba wiring!
from . import _wiring  # noqa: F401

from csr.constructors import *  # noqa: F401, F403
